from __future__ import annotations

from threading import Condition, Lock
from typing import Dict, List, Union

from weavel._request import (
    UnionRequest
)


class BufferStorage:
    def __init__(self, max_buffer_size: int):
        self.max_buffer_size = max_buffer_size
        self.buffer: List[
            UnionRequest
        ] = []
        self.buffer_lock = Lock()
        self.not_full_cv = Condition(self.buffer_lock)
        self.not_empty_cv = Condition(self.buffer_lock)

    @property
    def buffer_size(self) -> int:
        return len(self.buffer)

    def push(self, event: Dict) -> None:
        with self.buffer_lock:
            while self.buffer_size >= self.max_buffer_size:
                self.not_empty_cv.notify()  # immediately wake up possibly sleeping consumer
                self.not_full_cv.wait()
            self.buffer.append(event)
            if self.buffer_size >= self.max_buffer_size:
                self.not_empty_cv.notify()

    def pull(
        self, batch_size: int
    ) -> List[UnionRequest]:
        while not self.buffer_size:
            self.not_empty_cv.wait()
        pull_size = min(batch_size, self.buffer_size)
        out = self.buffer[:pull_size]
        self.buffer = self.buffer[pull_size:]
        self.not_full_cv.notify()
        return out

    def pull_all(
        self,
    ) -> List[UnionRequest]:
        out = self.buffer[:]
        self.buffer = []
        self.not_full_cv.notify()
        return out



import logging
import os
import sys
import typing as t

import structlog
from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

console = Console(width=200, theme=Theme({"logging.level": "bold"}))

LOGGER_NAME = "weavel"


logger: structlog.stdlib.BoundLogger = structlog.get_logger(LOGGER_NAME)


class LogSettings:
    def __init__(
        self, output_type: str, method: str, file_name: t.Optional[str]
    ) -> None:
        self.output_type = output_type
        self.method = method
        self.file_name = file_name
        self._configure_structlog()

    def _configure_structlog(self):
        if self.output_type == "str":
            renderer = self._create_rich_renderer()
            # renderer = structlog.dev.ConsoleRenderer()
        else:
            renderer = structlog.processors.JSONRenderer()

        structlog.configure(
            processors=[
                structlog.stdlib.add_logger_name,
                structlog.stdlib.add_log_level,
                structlog.processors.TimeStamper(fmt="iso"),
                structlog.processors.StackInfoRenderer(),
                structlog.processors.format_exc_info,
                structlog.processors.UnicodeDecoder(),
                renderer,
            ],
            logger_factory=structlog.stdlib.LoggerFactory(),
            wrapper_class=structlog.stdlib.BoundLogger,
            cache_logger_on_first_use=True,
        )

    def _create_rich_renderer(self):
        def rich_renderer(logger, name, event_dict):
            event = event_dict.pop("event", "")
            # Remove redundant information
            event_dict.pop("logger", None)

            return event

        return rich_renderer

    def set_log_output(
        self,
        method: t.Optional[str] = None,
        file_name: t.Optional[str] = None,
        output_type: t.Optional[str] = None,
    ) -> None:
        if method is not None and method not in ["console", "file"]:
            raise ValueError("method provided can only be 'console', 'file'")

        if method == "file" and file_name is None:
            raise ValueError("file_name must be provided when method = 'file'")

        if method is not None:
            self.method = method
            self.file_name = file_name

        if output_type is not None and output_type not in ["str", "json"]:
            raise ValueError("output_type provided can only be 'str', 'json'")

        if output_type is not None:
            self.output_type = output_type

        # Update Renderer
        self._configure_structlog()

        # Grab the weavel logger
        log = logging.getLogger(LOGGER_NAME)
        for handler in log.handlers[:]:
            log.removeHandler(handler)

        # Add new Handler
        if self.method == "file":
            assert self.file_name is not None
            file_handler = logging.FileHandler(self.file_name)
            file_formatter = logging.Formatter(
                fmt="%(asctime)s - %(levelname)s - %(message)s"
            )
            file_handler.setFormatter(file_formatter)
            log.addHandler(file_handler)
        else:
            rich_handler = RichHandler(
                console=console, rich_tracebacks=True, markup=True
            )
            rich_handler.setFormatter(logging.Formatter())
            log.addHandler(rich_handler)


DEFAULT_LOG_LEVEL = os.environ.get("WEAVEL_LOG_LEVEL", "WARNING").upper()


# Set Defaults
_logger = logging.getLogger(LOGGER_NAME)
_logger.handlers = [RichHandler(console=console, rich_tracebacks=True, markup=True)]
_logger.setLevel(DEFAULT_LOG_LEVEL)

settings = LogSettings(output_type="str", method="console", file_name=None)
set_log_output = settings.set_log_output



from h11 import Data
from pydantic import BaseModel as PydanticBaseModel, Field
from typing import Any, Dict, List, Optional, Literal, Union
from enum import Enum


class BaseModel(PydanticBaseModel):
    """Extended Pydantic BaseModel"""

    class Config:
        arbitrary_types_allowed = True


class IngestionType(str, Enum):
    CaptureSession = "capture-session"
    IdentifyUser = "identify-user"
    CaptureMessage = "capture-message"
    CaptureTrackEvent = "capture-track-event"
    CaptureTrace = "capture-trace"
    CaptureSpan = "capture-span"
    CaptureLog = "capture-log"
    CaptureGeneration = "capture-generation"
    CaptureTestObservation = "capture-test-observation"
    UpdateTrace = "update-trace"
    UpdateSpan = "update-span"
    UpdateGeneration = "update-generation"


class IngestionBody(BaseModel):
    type: IngestionType


class BaseRecordBody(BaseModel):
    """
    Represents the base record body.

    Attributes:
        session_id (str, optional): The unique identifier for the session. Optional.
        record_id (str, optional): The unique identifier for the record. Optional.
        created_at (str, optional): The datetime when the record was captured. Optional.
        metadata (Optional[Dict[str, Any]]): Additional metadata associated with the record. Optional.
        ref_record_id (str, optional): The record ID to reference. Optional.
    """

    session_id: Optional[str] = Field(
        default=None, description="The unique identifier for the record."
    )
    record_id: Optional[str] = Field(
        default=None,
        description="The unique identifier for the record. Optional.",
    )
    created_at: Optional[str] = Field(
        default=None,
        description="The datetime when the record was captured. Optional.",
    )
    metadata: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Additional metadata associated with the record. Optional.",
    )
    ref_record_id: Optional[str] = Field(
        default=None,
        description="The record ID to reference. Optional.",
    )


class BaseObservationBody(BaseModel):
    """
    Represents the base observation body.

    Attributes:
        record_id (str): The unique identifier for the record.
        observation_id (str, optional): The unique identifier for the observation. Optional.
        created_at (str, optional): The datetime when the observation was captured. Optional.
        name (str): The name of the observation.
        parent_observation_id (str, optional): The parent observation ID. Optional.
    """

    record_id: Optional[str] = Field(
        default=None, description="The unique identifier for the record."
    )
    observation_id: Optional[str] = Field(
        default=None,
        description="The unique identifier for the observation. Optional.",
    )
    created_at: Optional[str] = Field(
        default=None,
        description="The datetime when the observation was captured. Optional.",
    )
    name: Optional[str] = Field(
        default=None,
        description="The name of the observation. Optional.",
    )
    metadata: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Additional metadata associated with the observation. Optional.",
    )
    parent_observation_id: Optional[str] = Field(
        default=None,
        description="The parent observation ID. Optional.",
    )


class CaptureSessionBody(BaseModel):
    """
    Represents the request body for capturing a session.

    Attributes:
        user_id (str, optional): The unique identifier for the user.
        session_id (str, optional): The unique identifier for the session data.
        metadata (Dict[str, Any], Optional): Additional metadata associated with the session. Optional.
        created_at (str, optional): The datetime when the session was opened. Optional.
    """

    user_id: Optional[str] = Field(
        default=None, description="The unique identifier for the user."
    )
    session_id: Optional[str] = Field(
        default=None, description="The unique identifier for the session data."
    )
    metadata: Optional[Dict[str, str]] = Field(
        default=None,
        description="Additional metadata associated with the session. Optional.",
    )
    created_at: Optional[str] = Field(
        default=None,
        description="The datetime when the session was opened. Optional.",
    )


class CaptureRecordBody(BaseRecordBody):
    session_id: str


class CaptureObservationBody(BaseObservationBody):
    name: Optional[str] = Field(
        default=None,
        description="The name of the observation. Optional.",
    )


class CaptureMessageBody(CaptureRecordBody):
    """
    Represents a capture message body.

    Attributes:
        role (Literal["user", "assistant", "system"]): The role of the session, can be 'user', 'assistant', or 'system'.
        content (str): The content of the record.
    """

    role: Literal["user", "assistant", "system"] = Field(
        ...,
        description="The role of the session, 'user', 'assistant', 'system'.",
    )
    content: str = Field(..., description="The content of the record.")


class CaptureTrackEventBody(CaptureRecordBody):
    """
    Represents the body of a capture track event.

    Attributes:
        name (str): The name of the record. Optional.
        properties (Optional[Dict[str, Any]]): Additional properties associated with the record. Optional.
    """

    name: str = Field(
        ...,
        description="The name of the record. Optional.",
    )
    properties: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Additional properties associated with the record. Optional.",
    )


class CaptureTraceBody(CaptureRecordBody):
    """
    Represents the body of a capture trace record.

    Attributes:
        name (str): The name of the record. Optional.
    """

    name: str = Field(
        ...,
        description="The name of the trace record. Optional.",
    )
    inputs: Optional[Union[Dict[str, Any], List[Any], str]] = Field(
        default=None,
        description="The inputs of the trace. Optional.",
    )
    outputs: Optional[Union[Dict[str, Any], List[Any], str]] = Field(
        default=None,
        description="The outputs of the trace. Optional.",
    )


class CaptureSpanBody(CaptureObservationBody):
    """
    Represents the body of a capture span observation.

    Attributes:
        inputs (Optional[Union[Dict[str, Any], List[Any], str]]): The inputs of the generation. Optional.
        outputs (Optional[Union[Dict[str, Any], List[Any], str]]): The outputs of the generation. Optional.
    """

    inputs: Optional[Union[Dict[str, Any], List[Any], str]] = Field(
        default=None,
        description="The inputs of the generation. Optional.",
    )
    outputs: Optional[Union[Dict[str, Any], List[Any], str]] = Field(
        default=None,
        description="The outputs of the generation. Optional.",
    )


class CaptureTestObservationBody(CaptureObservationBody):
    test_uuid: str = Field(
        ...,
        description="The unique identifier for the test.",
    )
    dataset_item_uuid: str = Field(
        ...,
        description="The unique identifier for the dataset item.",
    )
    inputs: Optional[Union[Dict[str, Any], List[Any], str]] = Field(
        default=None,
        description="The inputs of the generation. Optional.",
    )
    outputs: Optional[Union[Dict[str, Any], List[Any], str]] = Field(
        default=None,
        description="The outputs of the generation. Optional.",
    )


class CaptureLogBody(CaptureObservationBody):
    """
    Represents the body of a capture log observation.

    Attributes:
        value (str, optional): The value of the observation. Optional.
    """

    value: Optional[str] = Field(
        default=None,
        description="The value of the observation. Optional.",
    )


class CaptureGenerationBody(CaptureObservationBody):
    """
    Represents the body of a capture generation.

    Attributes:
        inputs (Optional[Union[Dict[str, Any], List[Any], str]]): The inputs of the generation. Optional.
        outputs (Optional[Union[Dict[str, Any], List[Any], str]]): The outputs of the generation. Optional.
        prompt_name (Optional[str]): The name of the prompt. Optional.
    """

    inputs: Optional[Union[Dict[str, Any], List[Any], str]] = Field(
        default=None,
        description="The inputs of the generation. Optional.",
    )
    outputs: Optional[Union[Dict[str, Any], List[Any], str]] = Field(
        default=None,
        description="The outputs of the generation. Optional.",
    )
    messages: Optional[List[Dict[str, str]]] = Field(
        default=None,
        description="The messages of the generation. Optional.",
    )
    model: Optional[str] = Field(
        default=None,
        description="The model of the generation. Optional.",
    )
    latency: Optional[float] = Field(
        default=None,
        description="The latency of the generation. Optional.",
    )
    tokens: Optional[Dict[str, Any]] = Field(
        default=None,
        description="The tokens of the generation. Optional.",
    )
    cost: Optional[float] = Field(
        default=None,
        description="The cost of the generation. Optional.",
    )
    prompt_name: Optional[str] = Field(
        default=None,
        description="The name of the prompt. Optional.",
    )


class UpdateTraceBody(BaseRecordBody):
    record_id: str
    ended_at: Optional[str] = None
    inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None
    outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None
    metadata: Optional[Dict[str, Any]] = None


class UpdateSpanBody(BaseObservationBody):
    observation_id: str
    ended_at: Optional[str] = None
    inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None
    outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None
    metadata: Optional[Dict[str, Any]] = None


class UpdateGenerationBody(BaseObservationBody):
    observation_id: str
    ended_at: Optional[str] = None
    inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None
    outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None
    metadata: Optional[Dict[str, Any]] = None


class IdentifyUserBody(BaseModel):
    """
    Represents the body of a request to identify a user.

    Attributes:
        user_id (str): The unique identifier for the user.
        properties (Dict[str, Any], optional): Additional properties associated with the track event. Optional.
        created_at (Optional[str], optional): The datetime when the User is identified. Optional.
    """

    user_id: str = Field(..., description="The unique identifier for the user.")
    properties: Dict[str, Any] = Field(
        description="Additional properties associated with the track event. Optional.",
    )
    created_at: Optional[str] = Field(
        default=None,
        description="The datetime when the User is identified. Optional.",
    )



from .client import Weavel
from .wrapper import WeavelOpenAI, AsyncWeavelOpenAI
from .object_clients import SessionClient, TraceClient, SpanClient, GenerationClient

from .utils import *

__version___ = "1.10.0"



from typing import Dict, List, Optional, Any, Literal, Union
from uuid import uuid4
from datetime import datetime, timezone
from pydantic import Field

from weavel._request import BaseModel


class Observation(BaseModel):
    record_id: Optional[str] = None
    observation_id: str = Field(default_factory=lambda: str(uuid4()))
    created_at: Optional[datetime] = Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    name: Optional[str] = None
    metadata: Optional[Dict[str, str]] = None
    parent_observation_id: Optional[str] = None


class Span(Observation):
    type: Literal["span"] = "span"
    inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None
    outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None
    ended_at: Optional[datetime] = None


class Generation(Observation):
    type: Literal["generation"] = "generation"
    prompt_name: Optional[str] = None
    inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None
    outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None
    ended_at: Optional[datetime] = None
    model: Optional[str] = None
    latency: Optional[float] = None
    tokens: Optional[Dict[str, Any]] = None
    cost: Optional[float] = None


class Log(Observation):
    type: Literal["log"] = "log"
    value: Optional[str] = None



from __future__ import annotations

import time
from datetime import datetime, timezone
from typing import Any, Awaitable, Dict, List, Literal, Optional, Union
from threading import Thread
from concurrent.futures import Future, ThreadPoolExecutor
from pydantic import BaseModel
from openai.lib._parsing._completions import type_to_response_format_param

from weavel._request import (
    CaptureSessionRequest,
    IdentifyUserRequest,
    CaptureMessageRequest,
    CaptureTrackEventRequest,
    CaptureTraceRequest,
    CaptureSpanRequest,
    CaptureLogRequest,
    CaptureGenerationRequest,
    UpdateTraceRequest,
    UpdateSpanRequest,
    UpdateGenerationRequest,
    CaptureTestObservationRequest,
    UnionRequest,
)
from weavel._body import (
    CaptureSessionBody,
    CaptureMessageBody,
    CaptureTrackEventBody,
    CaptureTraceBody,
    CaptureSpanBody,
    CaptureLogBody,
    CaptureGenerationBody,
    UpdateTraceBody,
    UpdateSpanBody,
    UpdateGenerationBody,
    IdentifyUserBody,
    CaptureTestObservationBody,
)

from weavel._constants import ENDPOINT_URL
from weavel._buffer_storage import BufferStorage
from weavel._api_client import APIClient, AsyncAPIClient
from weavel.utils import logger
from weavel.types import WvDatasetItem, WvDataset, WvPrompt, WvPromptVersion
from ape.common.types import ResponseFormat


class Worker:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(Worker, cls).__new__(cls)
        return cls._instance

    def __init__(
        self,
        api_key: str,
        base_url: Optional[str] = None,
        max_retry: int = 3,
        flush_interval: int = 60,
        flush_batch_size: int = 20,
    ) -> None:
        if not hasattr(self, "is_initialized"):
            self.api_key = api_key
            self.endpoint = ENDPOINT_URL if not base_url else base_url
            self.endpoint += "/public/v2"
            self.max_retry = max_retry
            self.flush_interval = flush_interval
            self.flush_batch_size = flush_batch_size

            self.api_client = APIClient()
            self.async_api_client = AsyncAPIClient()

            self.api_pool = ThreadPoolExecutor(max_workers=1)
            self.buffer_storage = BufferStorage(max_buffer_size=1000)
            self._running = True
            self._thread = Thread(target=self.consume_buffer, daemon=True)
            self._thread.start()
            self.is_initialized = True
            self.testing = False

    def open_session(
        self,
        session_id: str,
        created_at: datetime,
        user_id: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> str:
        """Start the new session for user_id.

        Returns:
            The session id.
        """
        if not self.testing:
            if (
                created_at.tzinfo is None
                or created_at.tzinfo.utcoffset(created_at) is None
            ):
                created_at = created_at.astimezone(timezone.utc)
            else:
                created_at = created_at

            # add task to buffer
            request = CaptureSessionRequest(
                body=CaptureSessionBody(
                    user_id=user_id,
                    session_id=session_id,
                    created_at=created_at.isoformat(),
                    metadata=metadata,
                )
            )

            self.buffer_storage.push(request)

        return

    def identify_user(self, user_id: str, properties: Dict) -> None:
        """
        Identifies a user with the given properties.

        Args:
            user_id (str): The ID of the user.
            properties (Dict): The properties associated with the user.

        Returns:
            None
        """
        if not self.testing:
            request = IdentifyUserRequest(
                body=IdentifyUserBody(
                    user_id=user_id,
                    properties=properties,
                    created_at=datetime.now(timezone.utc).isoformat(),
                )
            )
            self.buffer_storage.push(request)

        return

    def capture_message(
        self,
        session_id: str,
        record_id: str,
        created_at: datetime,
        role: Literal["user", "assistant", "system"],
        content: str,
        metadata: Optional[Dict[str, Any]] = None,
        ref_record_id: Optional[str] = None,
    ):
        if not self.testing:
            if (
                created_at.tzinfo is None
                or created_at.tzinfo.utcoffset(created_at) is None
            ):
                created_at = created_at.astimezone(timezone.utc)
            else:
                created_at = created_at

            request = CaptureMessageRequest(
                body=CaptureMessageBody(
                    session_id=session_id,
                    record_id=record_id,
                    created_at=created_at.isoformat(),
                    role=role,
                    content=content,
                    metadata=metadata,
                    ref_record_id=ref_record_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_track_event(
        self,
        session_id: str,
        record_id: str,
        created_at: datetime,
        name: str,
        properties: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        ref_record_id: Optional[str] = None,
    ):
        if not self.testing:
            if (
                created_at.tzinfo is None
                or created_at.tzinfo.utcoffset(created_at) is None
            ):
                created_at = created_at.astimezone(timezone.utc)
            else:
                created_at = created_at

            request = CaptureTrackEventRequest(
                body=CaptureTrackEventBody(
                    session_id=session_id,
                    record_id=record_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    properties=properties,
                    metadata=metadata,
                    ref_record_id=ref_record_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_trace(
        self,
        session_id: str,
        record_id: str,
        created_at: datetime,
        name: str,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        ref_record_id: Optional[str] = None,
    ):
        if not self.testing:
            if (
                created_at.tzinfo is None
                or created_at.tzinfo.utcoffset(created_at) is None
            ):
                created_at = created_at.astimezone(timezone.utc)
            else:
                created_at = created_at

            request = CaptureTraceRequest(
                body=CaptureTraceBody(
                    session_id=session_id,
                    record_id=record_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    ref_record_id=ref_record_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def update_trace(
        self,
        record_id: str,
        ended_at: Optional[datetime] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        ref_record_id: Optional[str] = None,
    ):

        if not self.testing:
            request = UpdateTraceRequest(
                body=UpdateTraceBody(
                    record_id=record_id,
                    ended_at=ended_at.isoformat() if ended_at else None,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    ref_record_id=ref_record_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_span(
        self,
        record_id: str,
        observation_id: str,
        created_at: datetime,
        name: str,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
    ):
        if not self.testing:
            request = CaptureSpanRequest(
                body=CaptureSpanBody(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def update_span(
        self,
        observation_id: str,
        ended_at: Optional[datetime] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
    ):

        if not self.testing:
            request = UpdateSpanRequest(
                body=UpdateSpanBody(
                    observation_id=observation_id,
                    ended_at=ended_at.isoformat() if ended_at else None,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_log(
        self,
        record_id: str,
        observation_id: str,
        created_at: datetime,
        name: str,
        value: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
    ):
        if not self.testing:
            request = CaptureLogRequest(
                body=CaptureLogBody(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    value=value,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_generation(
        self,
        observation_id: str,
        created_at: datetime,
        name: Optional[str] = None,
        record_id: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        messages: Optional[List[Dict[str, str]]] = None,
        metadata: Optional[Dict[str, str]] = None,
        model: Optional[str] = None,
        latency: Optional[float] = None,
        tokens: Optional[Dict[str, Any]] = None,
        cost: Optional[float] = None,
        parent_observation_id: Optional[str] = None,
        prompt_name: Optional[str] = None,
    ):
        if not self.testing:
            request = CaptureGenerationRequest(
                body=CaptureGenerationBody(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    messages=messages,
                    model=model,
                    latency=latency,
                    tokens=tokens,
                    cost=cost,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                    prompt_name=prompt_name,
                )
            )
            self.buffer_storage.push(request)
        return

    async def acapture_generation(
        self,
        observation_id: str,
        created_at: datetime,
        name: str,
        record_id: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        messages: Optional[List[Dict[str, str]]] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
        prompt_name: Optional[str] = None,
        model: Optional[str] = None,
        latency: Optional[float] = None,
        tokens: Optional[Dict[str, Any]] = None,
        cost: Optional[float] = None,
    ):
        if not self.testing:
            request = CaptureGenerationRequest(
                body=CaptureGenerationBody(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at.isoformat(),
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    messages=messages,
                    metadata=metadata,
                    model=model,
                    latency=latency,
                    tokens=tokens,
                    cost=cost,
                    parent_observation_id=parent_observation_id,
                    prompt_name=prompt_name,
                )
            )
            self.buffer_storage.push(request)
        return

    def update_generation(
        self,
        observation_id: str,
        ended_at: Optional[datetime] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
        parent_observation_id: Optional[str] = None,
        prompt_name: Optional[str] = None,
    ):
        if not self.testing:
            request = UpdateGenerationRequest(
                body=UpdateGenerationBody(
                    observation_id=observation_id,
                    ended_at=ended_at.isoformat() if ended_at else None,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                )
            )
            self.buffer_storage.push(request)
        return

    def capture_test_observation(
        self,
        created_at: datetime,
        name: str,
        test_uuid: str,
        dataset_item_uuid: str,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, str]] = None,
    ):
        if self.testing:
            request = CaptureTestObservationRequest(
                body=CaptureTestObservationBody(
                    created_at=created_at.isoformat(),
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    test_uuid=test_uuid,
                    dataset_item_uuid=dataset_item_uuid,
                )
            )
            self.buffer_storage.push(request)
        return

    def create_dataset(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> WvDataset:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            "/datasets",
            method="POST",
            json={
                "name": name,
                "description": description,
            },
        )
        if response.status_code == 201:
            return WvDataset(**response.json())
        else:
            raise Exception(f"Failed to create dataset: {response.text}")

    async def acreate_dataset(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> Awaitable[WvDataset]:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            "/datasets",
            method="POST",
            json={
                "name": name,
                "description": description,
            },
        )
        if response:
            return WvDataset(**response.json())

    def create_dataset_items(
        self,
        dataset_name: str,
        items: List[WvDatasetItem],
    ) -> None:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            "/dataset_items/batch",
            method="POST",
            json={
                "dataset_name": dataset_name,
                "items": items,
            },
        )
        if response.status_code != 200:
            raise Exception(f"Failed to create dataset items: {response.text}")

    async def acreate_dataset_items(
        self,
        dataset_name: str,
        items: List[WvDatasetItem],
    ) -> None:
        await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            "/dataset_items/batch",
            method="POST",
            json={
                "dataset_name": dataset_name,
                "items": items,
            },
        )

    def fetch_dataset(
        self,
        name: str,
    ) -> WvDataset:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            f"/datasets/{name}",
            method="GET",
        )

        if response.status_code == 200:
            return WvDataset(**response.json())
        else:
            raise Exception(f"Failed to get dataset: {response.text}")

    async def afetch_dataset(
        self,
        name: str,
    ) -> WvDataset:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            f"/datasets/{name}",
            method="GET",
        )

        if response:
            return WvDataset(**response.json())
        
    def delete_dataset(self, name: str) -> None:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            f"/datasets/{name}",
            method="DELETE",
        )
        if response.status_code != 200:
            raise Exception(f"Failed to delete dataset: {response.text}")
    
    async def adelete_dataset(self, name: str) -> None:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            f"/datasets/{name}",
            method="DELETE",
        )
        if response.status_code != 200:
            raise Exception(f"Failed to delete dataset: {response.text}")

    def create_test(
        self,
        test_uuid: str,
        dataset_name: str,
        tags: Optional[List[str]] = None,
    ) -> None:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            "/tests",
            method="POST",
            json={
                "test_uuid": test_uuid,
                "dataset_name": dataset_name,
                "tags": tags,
            },
        )
        if response.status_code != 200:
            raise Exception(f"Failed to create test: {response.text}")

    # create, fetch, delete, list prompts
    def create_prompt(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> WvPrompt:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            "/prompts",
            method="POST",
            json={
                "name": name,
                "description": description,
            },
        )

        if response.status_code == 201:
            return WvPrompt(**response.json())
        else:
            raise Exception(f"Failed to create prompt: {response.text}")

    async def acreate_prompt(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> Awaitable[WvPrompt]:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            "/prompts",
            method="POST",
            json={
                "name": name,
                "description": description,
            },
        )
        if response:
            return WvPrompt(**response.json())

    def fetch_prompt(
        self,
        name: str,
    ) -> WvPrompt:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{name}",
            method="GET",
        )

        if response.status_code == 200:
            return WvPrompt(**response.json())
        else:
            raise Exception(f"Failed to get prompt: {response.text}")

    async def afetch_prompt(
        self,
        name: str,
    ) -> WvPrompt:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{name}",
            method="GET",
        )
        if response:
            return WvPrompt(**response.json())

    def delete_prompt(self, name: str) -> None:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{name}",
            method="DELETE",
        )
        if response.status_code != 200:
            raise Exception(f"Failed to delete prompt: {response.text}")

    async def adelete_prompt(self, name: str) -> None:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{name}",
            method="DELETE",
        )

    def list_prompts(self) -> List[WvPrompt]:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            "/prompts",
            method="GET",
        )

        if response.status_code == 200:
            return [WvPrompt(**prompt) for prompt in response.json()]
        else:
            raise Exception(f"Failed to list prompts: {response.text}")

    async def alist_prompts(self) -> List[WvPrompt]:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            "/prompts",
            method="GET",
        )

        if response:
            return [WvPrompt(**prompt) for prompt in response.json()]

    # create, fetch, delete, list prompt versions
    def create_prompt_version(
        self,
        prompt_name: str,
        messages: List[Dict[str, Any]],
        model: str = "gpt-4o-mini",
        temperature: float = 0.0,
        response_format: Optional[ResponseFormat] = None,
        input_vars: Optional[Dict[str, Any]] = None,
        output_vars: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{prompt_name}/versions",
            method="POST",
            json={
                "messages": messages,
                "model": model,
                "temperature": temperature,
                "response_format": (
                    type_to_response_format_param(response_format)
                    if isinstance(response_format, BaseModel)
                    else response_format
                ),
                "input_vars": input_vars,
                "output_vars": output_vars,
                "metadata": metadata,
            },
        )
        if not response.ok:
            raise Exception(f"Failed to create prompt version: {response.text}")

    async def acreate_prompt_version(
        self,
        prompt_name: str,
        messages: List[Dict[str, Any]],
        model: str = "gpt-4o-mini",
        temperature: float = 0.0,
        response_format: Optional[ResponseFormat] = None,
        input_vars: Optional[Dict[str, Any]] = None,
        output_vars: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> WvPromptVersion:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{prompt_name}/versions",
            method="POST",
            json={
                "messages": messages,
                "model": model,
                "temperature": temperature,
                "response_format": (
                    type_to_response_format_param(response_format)
                    if isinstance(response_format, BaseModel)
                    else response_format
                ),
                "input_vars": input_vars,
                "output_vars": output_vars,
                "metadata": metadata,
            },
        )

        return WvPromptVersion(**response.json())

    def fetch_prompt_version(
        self, prompt_name: str, version: Union[str, int]
    ) -> WvPromptVersion:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{prompt_name}/versions/{version}",
            method="GET",
        )
        if response.status_code == 200:
            return WvPromptVersion(**response.json())
        else:
            raise Exception(f"Failed to fetch prompt version: {response.text}")

    async def afetch_prompt_version(
        self, prompt_name: str, version: Union[str, int]
    ) -> WvPromptVersion:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{prompt_name}/versions/{version}",
            method="GET",
        )
        if response:
            return WvPromptVersion(**response.json())

    def delete_prompt_version(self, prompt_name: str, version: int) -> None:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{prompt_name}/versions/{version}",
            method="DELETE",
        )
        if response.status_code != 200:
            raise Exception(f"Failed to delete prompt version: {response.text}")

    async def adelete_prompt_version(self, prompt_name: str, version: int) -> None:
        await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{prompt_name}/versions/{version}",
            method="DELETE",
        )

    def list_prompt_versions(self, prompt_name: int) -> List[WvPromptVersion]:
        response = self.api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{prompt_name}/versions",
            method="GET",
        )

        if response.status_code == 200:
            return response.json()
        else:
            raise Exception(f"Failed to list prompt versions: {response.text}")

    async def alist_prompt_versions(self, prompt_name: str) -> List[WvPromptVersion]:
        response = await self.async_api_client.execute(
            self.api_key,
            self.endpoint,
            f"/prompts/{prompt_name}/versions",
            method="GET",
        )

        if response:
            return response.json()

    def send_requests(
        self,
        requests: List[UnionRequest],
    ):
        """
        Sends a batch of requests to the API endpoint.

        Args:
            requests (List[Union[OpenSessionBody, CaptureRecordBody, CaptureObservationBody]]):
                A list of requests to be sent to the API endpoint.

        Returns:
            None: This method does not return any value.

        Raises:
            Exception: If an error occurs while sending the requests.

        """
        # logger.info(requests)
        for attempt in range(self.max_retry):
            logger.debug(requests)
            try:
                response = self.api_client.execute(
                    self.api_key,
                    self.endpoint,
                    "/batch",
                    method="POST",
                    json={"batch": [request.model_dump() for request in requests]},
                    timeout=10,
                )
                if response.status_code == 200:
                    return
                # from rich import print

                # print({"batch": [request.model_dump() for request in requests]})
            except Exception as e:
                print(e)
                time.sleep(2**attempt)
                continue

    def consume_buffer(self) -> None:
        """
        Continuously consumes the buffer and sends requests to the API.

        This method runs in a loop until the `_running` flag is set to False. It checks the buffer size and waits for
        the specified flush interval if the buffer is empty. If the buffer has data, it pulls a batch of data from
        the buffer and sends requests to the API using the `send_requests` method. The completion of the API task is
        handled by the `_handle_api_task_completion` method.

        Raises:
            Exception: If an error occurs while consuming the buffer.

        """
        while self._running:
            try:
                with self.buffer_storage.buffer_lock:
                    if self.buffer_storage.buffer_size == 0:
                        self.buffer_storage.not_empty_cv.wait(self.flush_interval)
                        continue
                    out = self.buffer_storage.pull(self.flush_batch_size)
                    future = self.api_pool.submit(self.send_requests, out)
                    future.add_done_callback(self._handle_api_task_completion)
                    self.buffer_storage.not_empty_cv.wait(self.flush_interval)
            except Exception as e:
                print(e)

    def flush(self) -> None:
        """
        Flushes the buffer by sending the buffered requests.

        This method pulls all the requests from the buffer and sends them in batches
        of size `self.flush_batch_size` to the `send_requests` method.

        Raises:
            Exception: If an error occurs while flushing the buffer.
        """
        try:
            with self.buffer_storage.buffer_lock:
                if self.buffer_storage.buffer_size == 0:
                    return
                out = self.buffer_storage.pull_all()
                # for request chunks size of self.flush_batch_size
                for i in range(0, len(out), self.flush_batch_size):
                    request = out[i : i + self.flush_batch_size]
                    self.send_requests(request)
        except Exception as e:
            print(e)

    def stop(self) -> None:
        """
        Stops the worker thread and flushes any remaining data in the buffer.

        This method sets the `_running` flag to False, notifies the buffer storage
        that it is not empty, joins the worker thread, and then calls the `flush`
        method to ensure any remaining data in the buffer is processed.

        Returns:
            None
        """
        self._running = False
        with self.buffer_storage.buffer_lock:
            self.buffer_storage.not_empty_cv.notify()
        self._thread.join()
        self.flush()

    def _handle_api_task_completion(self, future: Future) -> None:
        try:
            future.result()
        except Exception as e:
            print(e)



from datetime import datetime
from typing import Any, List, Union, Literal

from openai import NotGiven
from ._body import (
    BaseModel,
    IngestionType,
    IngestionBody,
    CaptureSessionBody,
    CaptureMessageBody,
    CaptureTrackEventBody,
    CaptureTraceBody,
    CaptureSpanBody,
    CaptureLogBody,
    CaptureGenerationBody,
    UpdateTraceBody,
    UpdateSpanBody,
    UpdateGenerationBody,
    IdentifyUserBody,
    CaptureTestObservationBody,
)


class CaptureSessionRequest(IngestionBody):
    """
    Represents a request to capture a session.

    Attributes:
        type (Literal[IngestionType.OpenSession]): The type of ingestion, set to IngestionType.OpenSession.
        body (OpenSessionBody): The body of the open session request.
    """

    type: Literal[IngestionType.CaptureSession] = IngestionType.CaptureSession
    body: CaptureSessionBody


class CaptureMessageRequest(IngestionBody):
    """
    Represents a request to capture a message.

    Attributes:
        type (Literal[IngestionType.CaptureMessage]): The type of ingestion, which is set to IngestionType.CaptureMessage.
        body (CaptureMessageBody): The body of the capture message request.
    """

    type: Literal[IngestionType.CaptureMessage] = IngestionType.CaptureMessage
    body: CaptureMessageBody


class CaptureTrackEventRequest(IngestionBody):
    """
    Represents a request to capture a track event.

    Attributes:
        type (Literal[IngestionType.CaptureTrackEvent]): The type of ingestion, set to IngestionType.CaptureTrackEvent.
        body (CaptureTrackEventBody): The body of the capture track event request.
    """

    type: Literal[IngestionType.CaptureTrackEvent] = IngestionType.CaptureTrackEvent
    body: CaptureTrackEventBody


class CaptureTraceRequest(IngestionBody):
    """
    Represents a request to capture a trace.

    Attributes:
        type (Literal[IngestionType.CaptureTrace]): The type of ingestion, set to IngestionType.CaptureTrace.
        body (CaptureTraceBody): The body of the capture trace request.
    """

    type: Literal[IngestionType.CaptureTrace] = IngestionType.CaptureTrace
    body: CaptureTraceBody


class CaptureSpanRequest(IngestionBody):
    """
    Represents a request to capture a span.

    Attributes:
        type (Literal[IngestionType.CaptureSpan]): The type of ingestion, set to IngestionType.CaptureSpan.
        body (CaptureSpanBody): The body of the capture span request.
    """

    type: Literal[IngestionType.CaptureSpan] = IngestionType.CaptureSpan
    body: CaptureSpanBody


class CaptureLogRequest(IngestionBody):
    """
    Represents a request to capture log data.

    Attributes:
        type (Literal[IngestionType.CaptureLog]): The type of ingestion, set to IngestionType.CaptureLog.
        body (CaptureLogBody): The body of the capture log request.
    """

    type: Literal[IngestionType.CaptureLog] = IngestionType.CaptureLog
    body: CaptureLogBody


class CaptureGenerationRequest(IngestionBody):
    """
    Represents a request for capture generation.

    Attributes:
        type (Literal[IngestionType.CaptureGeneration]): The type of ingestion, which is set to IngestionType.CaptureGeneration.
        body (CaptureGenerationBody): The body of the capture generation request.
    """

    type: Literal[IngestionType.CaptureGeneration] = IngestionType.CaptureGeneration
    body: CaptureGenerationBody

    def serialize_output(self, output: Any) -> Any:
        if isinstance(output, NotGiven):
            return "NOT_GIVEN"
        if isinstance(output, (int, float, bool, str, type(None))):
            return output
        if isinstance(output, datetime):
            return output.isoformat()
        if isinstance(output, tuple):
            return {output[0]: self.serialize_output(output[1])}
        if hasattr(output, "__dict__"):
            return {
                k: self.serialize_output(v)
                for k, v in output.__dict__.items()
                if not k.startswith("_")
            }
        if isinstance(output, list):
            return [self.serialize_output(item) for item in output]
        if isinstance(output, dict):
            return {k: self.serialize_output(v) for k, v in output.items()}
        return str(output)

    def model_dump(self):
        inputs = self.body.inputs
        serialized_inputs = (
            inputs if isinstance(inputs, str) else self.serialize_output(inputs)
        )
        outputs = self.body.outputs
        serialized_outputs = (
            outputs if isinstance(outputs, str) else self.serialize_output(outputs)
        )

        return {
            "type": self.type,
            "body": {
                "record_id": self.body.record_id,
                "observation_id": self.body.observation_id,
                "created_at": self.body.created_at,
                "name": self.body.name,
                "parent_observation_id": self.body.parent_observation_id,
                "inputs": serialized_inputs,
                "outputs": serialized_outputs,
                "messages": self.body.messages,
                "metadata": self.body.metadata,
                "prompt_name": self.body.prompt_name,
                "model": self.body.model,
                "latency": self.body.latency,
                "tokens": self.body.tokens,
                "cost": self.body.cost,
            },
        }


class CaptureTestObservationRequest(IngestionBody):
    type: Literal[IngestionType.CaptureTestObservation] = (
        IngestionType.CaptureTestObservation
    )
    body: CaptureTestObservationBody


class UpdateTraceRequest(IngestionBody):
    type: Literal[IngestionType.UpdateTrace] = IngestionType.UpdateTrace
    body: UpdateTraceBody


class UpdateSpanRequest(IngestionBody):
    type: Literal[IngestionType.UpdateSpan] = IngestionType.UpdateSpan
    body: UpdateSpanBody


class IdentifyUserRequest(IngestionBody):
    """
    Represents a request to identify a user.

    Attributes:
        type (Literal[IngestionType.IdentifyUser]): The type of ingestion, which is set to IngestionType.IdentifyUser.
        body (IdentifyUserBody): The body of the identification request.
    """

    type: Literal[IngestionType.IdentifyUser] = IngestionType.IdentifyUser
    body: IdentifyUserBody


class UpdateGenerationRequest(IngestionBody):
    type: Literal[IngestionType.UpdateGeneration] = IngestionType.UpdateGeneration
    body: UpdateGenerationBody


class BatchRequest(BaseModel):
    batch: List[
        Union[
            CaptureSessionRequest,
            IdentifyUserRequest,
            CaptureMessageRequest,
            CaptureTrackEventRequest,
            CaptureTraceRequest,
            CaptureSpanRequest,
            CaptureLogRequest,
            CaptureGenerationRequest,
            UpdateTraceRequest,
            UpdateSpanRequest,
            UpdateGenerationRequest,
        ]
    ]


UnionRequest = Union[
    CaptureSessionRequest,
    IdentifyUserRequest,
    CaptureMessageRequest,
    CaptureTrackEventRequest,
    CaptureTraceRequest,
    CaptureSpanRequest,
    CaptureLogRequest,
    CaptureGenerationRequest,
    CaptureTestObservationRequest,
    UpdateTraceRequest,
    UpdateSpanRequest,
    UpdateGenerationRequest,
]



from .datasets import WvDataset, WvDatasetItem
from .prompts import WvPrompt, WvPromptVersion
from .observations import Observation, Span, Generation, Log
from .records import Record, Message, TrackEvent, Trace
from .session import Session

__all__ = [
    "WvDataset",
    "WvDatasetItem",
    "WvPrompt",
    "WvPromptVersion",
    "Observation",
    "Span",
    "Generation",
    "Log",
    "Record",
    "Message",
    "TrackEvent",
    "Trace",
    "Session",
]



from __future__ import annotations

import asyncio
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager
import contextvars
from itertools import islice
import os
from datetime import datetime, timezone
import time
from typing import Callable, Dict, List, Literal, Optional, Any, Union
from uuid import uuid4
from dotenv import load_dotenv
from pydantic import BaseModel
from openai.lib._parsing._completions import type_to_response_format_param

from ape.common import (
    Prompt,
    BaseGenerator,
    Generator,
    BaseMetric,
    BaseGlobalMetric,
    Evaluator,
)
from ape.common.types import (
    ResponseFormat,
    DatasetItem,
    MetricResult,
    GlobalMetricResult,
)
from ape.common.types.response_format import OpenAIResponseFormat
from ape.common.global_metric import AverageGlobalMetric

from weavel._constants import ENDPOINT_URL
from weavel._worker import Worker
from weavel.clients.websocket_client import WebsocketClient, websocket_handler
from weavel.object_clients import (
    GenerationClient,
    SessionClient,
    SpanClient,
    TraceClient,
)
from weavel.utils import logger
from weavel.types import WvDataset, WvDatasetItem, WvPrompt, WvPromptVersion
from weavel.types.websocket import (
    WsLocalEvaluateRequest,
    WsLocalEvaluateResponse,
    WsLocalGenerateRequest,
    WsLocalGlobalMetricRequest,
    WsLocalMetricRequest,
    WsLocalTask,
    WsServerOptimizeResponse,
    WsServerTask,
)

load_dotenv()


class Weavel:
    """Client for interacting with the Weavel service.

    This class provides methods for creating and managing traces, tracking user actions,
    and closing the client connection.

    Args:
        api_key (str, optional): The API key for authenticating with the Weavel service.
            If not provided, the API key will be retrieved from the environment variable
            WEAVEL_API_KEY.

    Attributes:
        api_key (str): The API key used for authentication.

    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        max_retry: Optional[int] = 3,
        flush_interval: Optional[int] = 60,
        flush_batch_size: Optional[int] = 20,
    ):
        self.api_key = api_key or os.getenv("WEAVEL_API_KEY")
        self.base_url = base_url or ENDPOINT_URL
        assert self.api_key is not None, "API key not provided."
        self._worker = Worker(
            self.api_key,
            base_url=base_url,
            max_retry=max_retry,
            flush_interval=flush_interval,
            flush_batch_size=flush_batch_size,
        )
        self.ws_client = WebsocketClient(api_key=api_key, base_url=base_url)
        self._generator_var: contextvars.ContextVar[Optional[BaseGenerator]] = (
            contextvars.ContextVar("generator")
        )
        self._evaluator_var: contextvars.ContextVar[Optional[Evaluator]] = (
            contextvars.ContextVar("evaluator")
        )
        self._trainset_var: contextvars.ContextVar[Optional[List[DatasetItem]]] = (
            contextvars.ContextVar("trainset")
        )
        self._metric_var: contextvars.ContextVar[Optional[BaseMetric]] = (
            contextvars.ContextVar("metric")
        )
        self._global_metric_var: contextvars.ContextVar[Optional[BaseGlobalMetric]] = (
            contextvars.ContextVar("global_metric")
        )
        self.ws_client.register_handlers(self)

        self.testing = False

    def session(
        self,
        user_id: Optional[str] = None,
        session_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> "SessionClient":
        """Create a new session for the specified user.

        Args:
            user_id (str): The user ID for the session.
            session_id (str, optional): The session ID. If not provided, a new session ID will be generated.
            created_at (datetime, optional): The created_at for the session. If not provided, the current time will be used.
            metadata (Dict[str, str], optional): Additional metadata for the session.

        Returns:
            SessionClient: The session object.

        """
        # if user_id is None and session_id is None:
        #     raise ValueError("user_id or session_id must be provided.")

        if session_id is None:
            session_id = str(uuid4())
        if created_at is None:
            created_at = datetime.now(timezone.utc)

        session = SessionClient(
            user_id=user_id,
            session_id=session_id,
            created_at=created_at,
            metadata=metadata,
            weavel_client=self._worker,
        )

        if self.testing:
            return session

        # if user_id is not None:
        self._worker.open_session(
            session_id=session_id,
            created_at=created_at,
            user_id=user_id,
            metadata=metadata,
        )
        return session

    def identify(self, user_id: str, properties: Dict[str, Any]):
        """Identify a user with the specified properties.

        You can save any user information you know about a user for user_id, such as their email, name, or phone number in dict format.
        Properties will be updated for every call.
        """
        if self.testing:
            return
        self._worker.identify_user(user_id, properties)

    def trace(
        self,
        session_id: Optional[str] = None,
        record_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        name: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        ref_record_id: Optional[str] = None,
    ) -> TraceClient:
        """Create a new trace record or fetch an existing one

        Args:
            session_id (str, optional): The session ID for the trace.
            record_id (str, optional): The record ID. If not provided, a new record ID will be generated.
            created_at (datetime, optional): The created_at for the trace. If not provided, the current time will be used.
            name (str, optional): The name of the trace.
            inputs (Dict[str, Any], optional): The inputs for the trace.
            outputs (Dict[str, Any], optional): The outputs for the trace.
            metadata (Dict[str, Any], optional): Additional metadata for the trace.
            ref_record_id (str, optional): The record ID to reference.

        """

        if session_id is None and record_id is None:
            raise ValueError("session_id or record_id must be provided.")

        if session_id is not None and name is None:
            raise ValueError(
                "If you want to create a new trace, you must provide a name."
            )

        if session_id is not None:
            # Create a new trace
            if record_id is None:
                record_id = str(uuid4())
            if created_at is None:
                created_at = datetime.now(timezone.utc)

            if not self.testing:
                self._worker.capture_trace(
                    session_id=session_id,
                    record_id=record_id,
                    created_at=created_at,
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    ref_record_id=ref_record_id,
                )
            return TraceClient(
                session_id=session_id,
                record_id=record_id,
                created_at=created_at,
                name=name,
                inputs=inputs,
                outputs=outputs,
                metadata=metadata,
                weavel_client=self._worker,
            )
        else:
            # Fetch an existing trace
            return TraceClient(record_id=record_id, weavel_client=self._worker)

    def generation(
        self,
        record_id: Optional[str] = None,
        observation_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        name: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        messages: Optional[List[Dict[str, str]]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        parent_observation_id: Optional[str] = None,
        prompt_name: Optional[str] = None,
        model: Optional[str] = None,
        latency: Optional[float] = None,
        tokens: Optional[Dict[str, Any]] = None,
        cost: Optional[float] = None,
    ) -> GenerationClient:
        # if record_id, observation_id, parent_observation_id
        # None, None, None -> ValueError
        # None, Value -> Fetch
        # Value -> Create
        # None, -, Value -> Create

        if not record_id and not observation_id and not parent_observation_id:
            if observation_id is None:
                observation_id = str(uuid4())
            if created_at is None:
                created_at = datetime.now(timezone.utc)
            self._worker.capture_generation(
                record_id=record_id,
                observation_id=observation_id,
                created_at=created_at,
                name=name,
                inputs=inputs,
                outputs=outputs,
                messages=messages,
                metadata=metadata,
                parent_observation_id=parent_observation_id,
                prompt_name=prompt_name,
                model=model,
                latency=latency,
                tokens=tokens,
                cost=cost,
            )

        if record_id is not None or (
            record_id is None and parent_observation_id is not None
        ):
            # Create a new generation
            if observation_id is None:
                observation_id = str(uuid4())
            if created_at is None:
                created_at = datetime.now(timezone.utc)

            if not self.testing:
                self._worker.capture_generation(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at,
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    messages=messages,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                    prompt_name=prompt_name,
                )
            return GenerationClient(
                record_id=record_id,
                observation_id=observation_id,
                created_at=created_at,
                name=name,
                inputs=inputs,
                outputs=outputs,
                messages=messages,
                metadata=metadata,
                parent_observation_id=parent_observation_id,
                prompt_name=prompt_name,
                model=model,
                latency=latency,
                tokens=tokens,
                cost=cost,
                weavel_client=self._worker,
            )
        else:
            # Fetch an existing generation
            return GenerationClient(
                observation_id=observation_id,
                weavel_client=self._worker,
            )

    def span(
        self,
        record_id: Optional[str] = None,
        observation_id: Optional[str] = None,
        created_at: Optional[datetime] = None,
        name: Optional[str] = None,
        inputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        outputs: Optional[Union[Dict[str, Any], List[Any], str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        parent_observation_id: Optional[str] = None,
    ) -> SpanClient:
        # if record_id, observation_id, parent_observation_id
        # None, None, None -> ValueError
        # None, Value -> Fetch
        # Value -> Create
        # None, -, Value -> Create

        if not record_id and not observation_id and not parent_observation_id:
            raise ValueError(
                "One of the record_id, observation_id, or parent_observation_id must be provided."
            )

        if (
            record_id or (record_id is None and parent_observation_id)
        ) and name is None:
            raise ValueError(
                "If you want to create a new span, you must provide a name."
            )

        if record_id is not None or (
            record_id is None and parent_observation_id is not None
        ):
            # Create a new span
            if observation_id is None:
                observation_id = str(uuid4())
            if created_at is None:
                created_at = datetime.now(timezone.utc)

            if not self.testing:
                self._worker.capture_span(
                    record_id=record_id,
                    observation_id=observation_id,
                    created_at=created_at,
                    name=name,
                    inputs=inputs,
                    outputs=outputs,
                    metadata=metadata,
                    parent_observation_id=parent_observation_id,
                )
            return SpanClient(
                record_id=record_id,
                observation_id=observation_id,
                created_at=created_at,
                name=name,
                inputs=inputs,
                outputs=outputs,
                metadata=metadata,
                parent_observation_id=parent_observation_id,
                weavel_client=self._worker,
            )
        else:
            # Fetch an existing span
            return SpanClient(observation_id=observation_id, weavel_client=self._worker)

    def track(
        self,
        session_id: str,
        name: str,
        properties: Optional[Dict[str, Any]] = None,
        created_at: Optional[datetime] = None,
        metadata: Optional[Dict[str, Any]] = None,
        record_id: Optional[str] = None,
        ref_record_id: Optional[str] = None,
    ) -> None:
        """Track an event for the specified session.

        Args:
            session_id (str): The session ID for the event.
            name (str): The name of the event.
            properties (Dict[str, Any], optional): Additional properties for the event.

        """
        if created_at is None:
            created_at = datetime.now(timezone.utc)

        if not self.testing:
            self._worker.capture_track_event(
                session_id=session_id,
                record_id=record_id,
                created_at=created_at,
                name=name,
                properties=properties,
                metadata=metadata,
                ref_record_id=ref_record_id,
            )

    def create_dataset(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> None:
        """Upload a dataset to the Weavel service.

        Args:
            name (str): The name of the dataset.
            description (str): The description of the dataset.
        """

        self._worker.create_dataset(
            name=name,
            description=description,
        )

    async def acreate_dataset(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> WvDataset:
        """Upload a dataset to the Weavel service.

        Args:
            name (str): The name of the dataset.
            description (str): The description of the dataset.
        """

        return await self._worker.acreate_dataset(
            name=name,
            description=description,
        )

    def get_dataset(self, name: str) -> WvDataset:
        """
        Retrieves a dataset with the given name.

        Args:
            name (str): The name of the dataset to retrieve.

        Returns:
            Dataset: The retrieved dataset.
        """

        return self._worker.fetch_dataset(name)

    async def aget_dataset(self, name: str) -> WvDataset:
        """
        Retrieves a dataset with the given name.

        Args:
            name (str): The name of the dataset to retrieve.

        Returns:
            Dataset: The retrieved dataset.
        """

        return await self._worker.afetch_dataset(name)

    def delete_dataset(self, name: str) -> None:
        """Delete a dataset with the given name.

        Args:
            name (str): The name of the dataset to delete.
        """

        self._worker.delete_dataset(name)

    async def adelete_dataset(self, name: str) -> None:
        """Delete a dataset with the given name.

        Args:
            name (str): The name of the dataset to delete.
        """

        await self._worker.adelete_dataset(name)

    # create, fetch, delete, and list prompts
    def create_prompt(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> WvPrompt:
        """Upload a prompt to the Weavel service.

        Args:
            name (str): The name of the prompt.
            description (str): The description of the prompt.
        """

        return self._worker.create_prompt(
            name=name,
            description=description,
        )

    async def acreate_prompt(
        self,
        name: str,
        description: Optional[str] = None,
    ) -> WvPrompt:
        """Upload a prompt to the Weavel service.

        Args:
            name (str): The name of the prompt.
            description (str): The description of the prompt.
        """

        return await self._worker.acreate_prompt(
            name=name,
            description=description,
        )

    def fetch_prompt(self, name: str) -> WvPrompt:
        """
        Retrieves a prompt with the given name.

        Args:
            name (str): The name of the prompt to retrieve.

        Returns:
            Prompt: The retrieved prompt.
        """

        return self._worker.fetch_prompt(name)

    async def afetch_prompt(self, name: str) -> WvPrompt:
        """
        Retrieves a prompt with the given name.

        Args:
            name (str): The name of the prompt to retrieve.

        Returns:
            Prompt: The retrieved prompt.
        """

        return await self._worker.afetch_prompt(name)

    def delete_prompt(self, name: str) -> None:
        """Delete a prompt from the Weavel service.

        Args:
            name (str): The name of the prompt to delete.
        """

        self._worker.delete_prompt(name=name)

    async def adelete_prompt(self, name: str) -> None:
        """Delete a prompt from the Weavel service asynchronously.

        Args:
            name (str): The name of the prompt to delete.
        """

        await self._worker.adelete_prompt(name=name)

    def list_prompts(self) -> List[WvPrompt]:
        """List all prompts that user created.

        Returns:
            List[Prompt]: A list of all prompts.
        """

        return self._worker.list_prompts()

    async def alist_prompts(self) -> List[WvPrompt]:
        """List all prompts that user created.

        Returns:
            List[Prompt]: A list of all prompts.
        """

        return await self._worker.alist_prompts()

    # create, fetch, delete, and list prompt versions
    def create_prompt_version(
        self,
        prompt_name: str,
        messages: List[Dict[str, Any]],
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        response_format: Optional[ResponseFormat] = None,
        input_vars: Optional[Dict[str, Any]] = None,
        output_vars: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Create a new version of a prompt.

        Args:
            prompt_name (str): The name of the prompt.
            messages (List[Dict[str, Any]]): The messages for the prompt version.
            model (Optional[str]): The model to use for this prompt version. Default is 'gpt-4o-mini'.
            temperature (Optional[float]): The temperature setting for the model. Default is 0.0.
            response_format (Optional[ResponseFormat]): The response format for the prompt.
            input_vars (Optional[Dict[str, Any]]): The input variables for the prompt.
            output_vars (Optional[Dict[str, Any]]): The output variables for the prompt.
            metadata (Optional[Dict[str, Any]]): Additional metadata for the prompt version.
        """
        self._worker.create_prompt_version(
            prompt_name=prompt_name,
            messages=messages,
            model=model,
            temperature=temperature,
            response_format=response_format,
            input_vars=input_vars,
            output_vars=output_vars,
            metadata=metadata,
        )

    async def acreate_prompt_version(
        self,
        prompt_name: str,
        messages: List[Dict[str, Any]],
        model: Optional[str] = None,
        temperature: Optional[float] = None,
        response_format: Optional[OpenAIResponseFormat] = None,
        input_vars: Optional[Dict[str, Any]] = None,
        output_vars: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> WvPromptVersion:
        """Create a new version of a prompt asynchronously.

        Args:
            prompt_name (str): The name of the prompt.
            messages (List[Dict[str, Any]]): The messages for the prompt version.
            model (Optional[str]): The model to use for this prompt version. Default is 'gpt-4o-mini'.
            temperature (Optional[float]): The temperature setting for the model. Default is 0.0.
            response_format (Optional[Dict[str, Any]]): The response format for the prompt.
            input_vars (Optional[Dict[str, Any]]): The input variables for the prompt.
            output_vars (Optional[Dict[str, Any]]): The output variables for the prompt.
            metadata (Optional[Dict[str, Any]]): Additional metadata for the prompt version.
        """
        return await self._worker.acreate_prompt_version(
            prompt_name=prompt_name,
            messages=messages,
            model=model,
            temperature=temperature,
            response_format=response_format,
            input_vars=input_vars,
            output_vars=output_vars,
            metadata=metadata,
        )

    def fetch_prompt_version(
        self, prompt_name: str, version: Union[str, int]
    ) -> WvPromptVersion:
        """Fetch a specific version of a prompt.

        Args:
            prompt_name (str): The name of the prompt.
            version (Union[str, int]): The version identifier to fetch. Get latest version by version = 'latest'. Otherwise, specify the version number.

        Returns:
            PromptVersion: The prompt version details.
        """
        return self._worker.fetch_prompt_version(
            prompt_name=prompt_name, version=version
        )

    async def afetch_prompt_version(
        self, prompt_name: str, version: Union[str, int]
    ) -> WvPromptVersion:
        """Fetch a specific version of a prompt asynchronously.
        Get latest version by version = 'latest'. Otherwise, specify the version number.

        Args:
            prompt_name (str): The name of the prompt.
            version (str): The version identifier to fetch.

        Returns:
            PromptVersion: The prompt version details.
        """

        return await self._worker.afetch_prompt_version(
            prompt_name=prompt_name, version=version
        )

    def list_prompt_versions(self, prompt_name: str) -> List[WvPromptVersion]:
        """List all versions of a specific prompt.

        Args:
            prompt_name (str): The name of the prompt.

        Returns:
            List[PromptVersion]: A list of version identifiers.
        """
        return self._worker.list_prompt_versions(prompt_name=prompt_name)

    async def alist_prompt_versions(self, prompt_name: str) -> List[WvPromptVersion]:
        """List all versions of a specific prompt asynchronously.

        Args:
            prompt_name (str): The name of the prompt.

        Returns:
            List[PromptVersion]: A list of version identifiers.
        """

        return await self._worker.alist_prompt_versions(prompt_name=prompt_name)

    def delete_prompt_version(self, prompt_name: str, version: str) -> None:
        """Delete a specific version of a prompt.

        Args:
            prompt_name (str): The name of the prompt.
            version (str): The version identifier to delete.
        """

        self._worker.delete_prompt_version(prompt_name=prompt_name, version=version)

    async def adelete_prompt_version(self, prompt_name: str, version: str) -> None:
        """Delete a specific version of a prompt asynchronously.

        Args:
            prompt_name (str): The name of the prompt.
            version (str): The version identifier to delete.
        """

        await self._worker.adelete_prompt_version(
            prompt_name=prompt_name, version=version
        )

    def create_dataset_items(
        self,
        dataset_name: str,
        items: Union[List[Dict[str, Any]], List[WvDatasetItem]],
    ) -> None:
        """Upload dataset items to the Weavel service.

        Args:
            dataset_name (str): The name of the dataset.
            items (Union[List[Dict[str, Any]], List[DatasetItem]]): The dataset items to upload.
        """

        for item in items:
            if isinstance(item, WvDatasetItem):
                item = item.model_dump()

        self._worker.create_dataset_items(dataset_name, items)

    async def acreate_dataset_items(
        self,
        dataset_name: str,
        items: Union[List[DatasetItem], List[WvDatasetItem]],
    ) -> None:
        """Upload dataset items to the Weavel service.

        Args:
            dataset_name (str): The name of the dataset.
            items (Union[List[Dict[str, Any]], List[DatasetItem]]): The dataset items to upload.
        """

        await self._worker.acreate_dataset_items(
            dataset_name,
            [item.model_dump() for item in items if isinstance(item, WvDatasetItem)],
        )

    def test(
        self,
        func: Callable,
        dataset_name: str,
        batch_size: int = 50,
        delay: int = 10,
        tags: Optional[List[str]] = None,
    ) -> None:
        """Test the function with the dataset.
        It must be used in the jupyter notebook environment.

        Args:
            func (callable): The function to test.
            dataset_name (str): The name of the dataset.
            tags: (List[str], optional): The tags for the test.
            batch_size (int, optional): The batch size for the test. Default is 50.
            delay (int, optional): The delay (in seconds) between each batch. Default is 10.
            mock_inputs (Dict[str, Any], optional): The mock input variables to run the function.
        """
        import nest_asyncio

        nest_asyncio.apply()

        # fetch dataset from server
        dataset: WvDataset = self._worker.fetch_dataset(dataset_name)

        # create test instance in database
        test_uuid = str(uuid4())
        self._worker.create_test(test_uuid, dataset_name, tags)

        self.testing = True
        self._worker.testing = True

        def _runner(
            func: callable,
            inputs: Union[Dict[str, Any], List[Any], str],
            dataset_item_uuid: str,
            test_uuid: str,
        ):
            # run the function and capture the result
            if isinstance(inputs, str):
                result = func(inputs)
            else:
                result = func(**inputs)

            if not isinstance(result, dict) and not isinstance(result, str):
                result = str(result)

            self._worker.capture_test_observation(
                created_at=datetime.now(timezone.utc),
                name=dataset_name,
                test_uuid=test_uuid,
                dataset_item_uuid=dataset_item_uuid,
                inputs=inputs,
                outputs=result,
            )

        async def _arunner(
            func: Callable,
            inputs: Union[Dict[str, Any], List[Any], str],
            dataset_item_uuid: str,
            test_uuid: str,
        ):
            # if "_RAW_VALUE_" in inputs:
            if isinstance(inputs, str):
                result = await func(inputs)
            else:
                result = await func(**inputs)

            if not isinstance(result, dict) and not isinstance(result, str):
                result = str(result)

            self._worker.capture_test_observation(
                created_at=datetime.now(timezone.utc),
                name=dataset_name,
                test_uuid=test_uuid,
                dataset_item_uuid=dataset_item_uuid,
                inputs=inputs,
                outputs=result,
            )

        async def run_async(func: Callable, dataset_items: List[WvDatasetItem]):
            for i in range(0, len(dataset_items), batch_size):
                batch = list(islice(dataset_items, i, i + batch_size))
                coros = [
                    _arunner(func, data.inputs, data.uuid, test_uuid) for data in batch
                ]
                await asyncio.gather(*coros)
                if i + batch_size < len(dataset_items):
                    await asyncio.sleep(delay)

        def run_threaded(func, dataset_items: List[WvDatasetItem]):
            with ThreadPoolExecutor() as executor:
                for i in range(0, len(dataset_items), batch_size):
                    batch = list(islice(dataset_items, i, i + batch_size))
                    futures = [
                        executor.submit(
                            _runner, func, data.inputs, data.uuid, test_uuid
                        )
                        for data in batch
                    ]
                    for future in futures:
                        future.result()
                    if i + batch_size < len(dataset_items):
                        time.sleep(delay)

        print("Testing...")
        if asyncio.iscoroutinefunction(func):
            asyncio.run(run_async(func, dataset.items))
        else:
            run_threaded(func, dataset.items)

        print("Test finished. Start Logging...")

        self._worker.testing = False
        self.testing = False

        self._worker.flush()
        print("Test completed.")

    def close(self):
        """Close the client connection."""
        self._worker.stop()

    def flush(self):
        """Flush the buffer."""
        self._worker.flush()

    @contextmanager
    def _ape_context(
        self,
        generator: Optional[BaseGenerator],
        evaluator: Optional[Evaluator],
        metric: Optional[BaseMetric],
        trainset: Optional[List[DatasetItem]],
        global_metric: Optional[BaseGlobalMetric] = AverageGlobalMetric(),
    ):
        token_generator = self._generator_var.set(generator)
        token_evaluator = self._evaluator_var.set(evaluator)
        token_trainset = self._trainset_var.set(trainset)
        token_metric = self._metric_var.set(metric)
        token_global_metric = self._global_metric_var.set(global_metric)
        try:
            yield
        finally:
            self._generator_var.reset(token_generator)
            self._evaluator_var.reset(token_evaluator)
            self._trainset_var.reset(token_trainset)
            self._metric_var.reset(token_metric)
            self._global_metric_var.reset(token_global_metric)

    def _get_generator(self) -> Optional[BaseGenerator]:
        return self._generator_var.get()

    def _get_evaluator(self) -> Optional[Evaluator]:
        return self._evaluator_var.get()

    def _get_trainset(self) -> Optional[List[DatasetItem]]:
        return self._trainset_var.get()

    def _get_metric(self) -> Optional[BaseMetric]:
        return self._metric_var.get()

    def _get_global_metric(self) -> Optional[BaseGlobalMetric]:
        return self._global_metric_var.get()

    def _set_global_metric(self, global_metric: Optional[BaseGlobalMetric]):
        self._global_metric_var.set(global_metric)

    @websocket_handler(WsLocalTask.GENERATE.value)
    async def handle_generation_request(self, data: WsLocalGenerateRequest):
        logger.debug("Handling generation request...")
        generator = self._get_generator()
        if not generator:
            raise AttributeError("Generate not set")
        return await generator(prompt=Prompt(**data["prompt"]), inputs=data["inputs"])

    @websocket_handler(WsLocalTask.EVALUATE.value)
    async def handle_evaluation_request(
        self, data: WsLocalEvaluateRequest
    ) -> WsLocalEvaluateResponse:
        logger.debug("Handling evaluation request...")
        evaluator = self._get_evaluator()
        trainset = self._get_trainset()
        if not evaluator:
            raise AttributeError("Evaluation not set")
        return_only_score = data.get("return_only_score", True)

        logger.debug(f"Evaluating {len(trainset)} items")
        logger.debug(f"Return Only Score : {return_only_score}")
        if return_only_score:
            score = await evaluator(
                prompt=Prompt(**data["prompt"]),
                testset=trainset,
                return_only_score=True,
            )
            return {
                "score": score,
            }
        else:
            preds, eval_results, global_result = await evaluator(
                prompt=Prompt(**data["prompt"]),
                testset=trainset,
                return_only_score=False,
            )
            return {
                "preds": preds,
                "eval_results": [i.model_dump() for i in eval_results],
                "global_result": global_result.model_dump(),
            }

    @websocket_handler(WsLocalTask.METRIC.value)
    async def handle_metric_request(self, data: WsLocalMetricRequest):
        logger.debug("Handling metric request...")
        metric = self._get_metric()
        if not metric:
            raise AttributeError("Metric not set")
        res = await metric(dataset_item=data["dataset_item"], pred=data["pred"])
        return res.model_dump()

    @websocket_handler(WsLocalTask.GLOBAL_METRIC.value)
    async def handle_global_metric_request(self, data: WsLocalGlobalMetricRequest):
        logger.debug("Handling global metric request...")
        global_metric = self._get_global_metric()
        if not global_metric:
            raise AttributeError("Global Metric not set")
        results = [MetricResult(**r) for r in data["results"]]
        res = await global_metric(results=results)
        return res.model_dump()

    @websocket_handler(WsServerTask.OPTIMIZE.value)
    async def handle_optimization_result(self, data: Dict[str, Any]):
        # Extract the correlation_id from the response data
        correlation_id = data.get("correlation_id")
        if not correlation_id:
            logger.error("No correlation_id found in the OPTIMIZE_response")
            return

        optimization_result = data.get("result")
        if optimization_result is None:
            logger.error("No 'result' field found in the OPTIMIZE_response")
            return

        # Put the result in the appropriate response queue
        await self.ws_client._responses[correlation_id].put(optimization_result)

        # Set the event to notify that the response has been received
        event = self.ws_client._pending_requests.get(correlation_id)
        if event:
            event.set()
        else:
            logger.error(
                f"No pending request found for correlation_id: {correlation_id}"
            )

    # Optimization
    async def optimize(
        self,
        base_prompt: Prompt | WvPromptVersion,
        models: List[str],
        metric: BaseMetric,
        trainset: List[DatasetItem] | WvDataset,
        generator: Optional[BaseGenerator] = Generator(),
        global_metric: Optional[BaseGlobalMetric] = None,
        algorithm: Literal[
            "dspy_mipro", "few_shot", "text_gradient", "optuna", "expel"
        ] = "dspy_mipro",
        # DspyMiproTrainer & OptunaTrainer & FewShotTrainer params
        num_candidates: Optional[int] = 10,
        # DspyMiproTrainer & FewShotTrainer params
        max_bootstrapped_demos: Optional[int] = 5,
        max_labeled_demos: Optional[int] = 5,
        success_score: Optional[float] = 1.0,
        # DspyMiproTrainer & OptunaTrainer params
        minibatch_size: Optional[int] = 25,
        max_steps: Optional[int] = 20,
        # TextGradientTrainer params
        batch_size: Optional[int] = 4,
        early_stopping_rounds: Optional[int] = 10,
        validation_type: Optional[Literal["trainset", "valset", "all"]] = "trainset",
        # TextGradientTrainer & ExpelTrainer params
        max_proposals_per_step: Optional[int] = 5,
        # ExpelTrainer params
        target_subgroup: Literal["success", "failure", "all"] = "all",
    ) -> Prompt:
        # Set or create base prompt
        if isinstance(base_prompt, Prompt):
            wv_prompt = None
            if base_prompt.name:
                try:
                    wv_prompt = await self.afetch_prompt(name=base_prompt.name)
                except Exception:
                    pass
            if not wv_prompt:
                wv_prompt = await self.acreate_prompt(
                    name=base_prompt.name or str(uuid4())
                )
            wv_prompt_version = await self.acreate_prompt_version(
                prompt_name=wv_prompt.name,
                messages=base_prompt.messages,
                model=base_prompt.model,
                temperature=base_prompt.temperature,
                response_format=(
                    type_to_response_format_param(base_prompt.response_format)
                    if isinstance(base_prompt.response_format, BaseModel)
                    else base_prompt.response_format
                ),
                input_vars=base_prompt.inputs_desc,
                output_vars=base_prompt.outputs_desc,
                metadata=base_prompt.metadata,
            )
        elif isinstance(base_prompt, WvPromptVersion):
            wv_prompt_version = base_prompt
        else:
            raise ValueError(
                "base_prompt must be either a Prompt or WvPromptVersion object"
            )

        dataset_created = False
        if not isinstance(trainset, WvDataset):
            dataset_name = f"trainset-{uuid4()}"
            dataset = await self.acreate_dataset(name=dataset_name)
            dataset_created = True
            dataset_items = [
                WvDatasetItem(inputs=item["inputs"], outputs=item["outputs"])
                for item in trainset
            ]
            await self.acreate_dataset_items(dataset_name, dataset_items)
        else:
            if not trainset.items:
                dataset = await self.aget_dataset(trainset.name)
            else:
                dataset = trainset
            dataset_items = trainset.items

        trainset = [
            DatasetItem(inputs=d.inputs, outputs=d.outputs) for d in dataset_items
        ]

        evaluator = Evaluator(
            metric=metric,
            global_metric=global_metric,
            testset=trainset,
        )

        try:
            with self._ape_context(
                generator=generator,
                evaluator=evaluator,
                metric=metric,
                trainset=trainset,
            ):
                async with self.ws_client:
                    res = await self.ws_client.request(
                        type=WsServerTask.OPTIMIZE,
                        data={
                            "base_prompt_version_uuid": wv_prompt_version.uuid,
                            "models": models,
                            "trainset_uuid": dataset.uuid,
                            "algorithm": algorithm,
                            "num_candidates": num_candidates,
                            "max_bootstrapped_demos": max_bootstrapped_demos,
                            "max_labeled_demos": max_labeled_demos,
                            "success_score": success_score,
                            "minibatch_size": minibatch_size,
                            "max_steps": max_steps,
                            "batch_size": batch_size,
                            "early_stopping_rounds": early_stopping_rounds,
                            "validation_type": validation_type,
                            "max_proposals_per_step": max_proposals_per_step,
                            "target_subgroup": target_subgroup,
                        },
                    )

                    logger.info("Optimization complete!")
                    logger.info(f"View all prompts at: {res['url']}")
                    return Prompt(**res["optimized_prompt"])
        except Exception as exc:
            raise exc
        finally:
            if dataset_created:
                await self.adelete_dataset(dataset.name)



from enum import StrEnum
from typing import Any, Dict, Iterable, List, Optional, Union
from typing_extensions import TypedDict
from openai.types.chat.completion_create_params import ChatCompletionMessageParam
from ape.common.types import DatasetItem, MetricResult, GlobalMetricResult
class WsLocalTask(StrEnum):
    GENERATE = "GENERATE"
    EVALUATE = "EVALUATE"
    METRIC = "METRIC"
    GLOBAL_METRIC = "GLOBAL_METRIC"


class WsServerTask(StrEnum):
    OPTIMIZE = "OPTIMIZE"


class WsServerOptimizeResponse(StrEnum):
    OPTIMIZATION_COMPLETE = "OPTIMIZATION_COMPLETE"


class BaseWsLocalRequest(TypedDict):
    type: WsLocalTask
    correlation_id: str


class WsLocalGenerateRequest(BaseWsLocalRequest):
    prompt: Dict[str, Any]
    inputs: Dict[str, Any]


class WsLocalEvaluateRequest(BaseWsLocalRequest):
    prompt: Dict[str, Any]


class WsLocalEvaluateResponse(TypedDict):
    score: Optional[float]
    preds: Optional[List[Union[str, Dict[str, Any]]]]
    eval_results: Optional[List[MetricResult]]
    global_result: Optional[GlobalMetricResult]

class WsLocalGlobalMetricRequest(BaseWsLocalRequest):
    results: List[MetricResult]
    
class WsLocalGlobalMetricResponse(TypedDict):
    global_result: GlobalMetricResult

class WsLocalMetricRequest(BaseWsLocalRequest):
    dataset_item: DatasetItem
    pred: Union[str, Dict[str, Any]]



from typing import Any, Dict, List, Optional, Union, Literal
from pydantic import BaseModel
from ape.common.types import ResponseFormat


class WvPrompt(BaseModel):
    uuid: Optional[str] = None
    name: str
    description: Optional[str] = None
    created_at: str


class WvPromptVersion(BaseModel):
    uuid: Optional[str] = None
    prompt_uuid: Optional[str] = None
    version: int
    messages: List[Dict[str, Any]]
    model: str
    temperature: float
    response_format: Optional[ResponseFormat] = None
    input_vars: Optional[Dict[str, Any]] = None
    output_vars: Optional[Dict[str, Any]] = None
    metadata: Optional[Dict[str, Any]] = None
    created_at: str



import asyncio
from functools import wraps
import json
import datetime
import os
import re
import textwrap
import time  # Add this import

from uuid import UUID, uuid4
from typing import Callable, Dict, Any, Optional, AsyncGenerator, List
from dotenv import load_dotenv
from collections import defaultdict
from asyncio import Queue, Semaphore

from websockets.client import connect, WebSocketClientProtocol
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
from readerwriterlock import rwlock
from tenacity import retry, stop_after_attempt, wait_fixed, retry_if_exception_type

from weavel.utils.logging import logger
from weavel._constants import ENDPOINT_URL
from weavel.types.websocket import WsLocalTask, WsServerTask


load_dotenv()


class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, datetime.datetime):
            aware_datetime = obj.replace(tzinfo=datetime.timezone.utc)
            return aware_datetime.isoformat()  # This will include timezone information
        return super().default(obj)


def websocket_handler(message_type: str):
    def decorator(func: Callable):
        @wraps(func)
        async def wrapper(self, data: Dict[str, Any]):
            try:
                result = await func(self, data)
                if result is not None:
                    response = {
                        "type": f"{message_type}_response",
                        "correlation_id": data.get("correlation_id"),
                        "data": result,
                    }
                    logger.debug(f"Sending response for {message_type}: {response}")
                    await self.ws_client.send_message(response)
            except Exception as e:
                logger.exception(f"Error in {func.__name__}")

        wrapper.message_type = message_type
        return wrapper

    return decorator


class WebsocketClient:
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(WebsocketClient, cls).__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    def __init__(
        self,
        api_key: str,
        base_url: Optional[str] = None,
    ):
        if not hasattr(self, "_initialized"):
            return
        self._initialized = True
        self.api_key = api_key
        protocol = "ws" if os.getenv("WEAVEL_TESTMODE") == "true" else "wss"
        base_url = base_url or ENDPOINT_URL
        domain = base_url.split("://")[1]
        self.endpoint = f"{protocol}://{domain}/public/v2/ws/open"
        self.ws: Optional[WebSocketClientProtocol] = None
        self.message_handlers: Dict[str, Callable] = {}
        self.heartbeat_task = None
        self.message_handler_task = None

        # Correlation Management
        self._pending_requests: Dict[str, asyncio.Event] = {}
        self._responses: Dict[str, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._lock = asyncio.Lock()  # To ensure thread-safe operations
        self._timeout_reset_events: Dict[str, asyncio.Event] = defaultdict(
            asyncio.Event
        )

        self._semaphore = Semaphore(100)  # Limit to 100 concurrent tasks

    async def __aenter__(self):
        await self.connect_to_gateway()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close_connection()

    @retry(
        stop=stop_after_attempt(12 * 24),
        wait=wait_fixed(5 * 60),
        retry=retry_if_exception_type(
            (ConnectionClosedError, ConnectionClosedOK, TimeoutError)
        ),
    )
    async def connect_to_gateway(self):
        headers = {
            "Authorization": f"Bearer {self.api_key}",
        }
        try:
            self.ws = await connect(
                self.endpoint,
                extra_headers=headers,
                ping_interval=30,
                ping_timeout=10,
                close_timeout=60,
            )
            logger.info("WebSocket connection established")
            self.heartbeat_task = asyncio.create_task(self.heartbeat())
            self.message_handler_task = asyncio.create_task(self.message_handler())
        except Exception as error:
            logger.exception("Error connecting to the gateway")
            raise

    async def close_connection(self):
        if self.ws and not self.ws.closed:
            await self.ws.close()
            logger.info("WebSocket connection closed")
        if self.heartbeat_task:
            self.heartbeat_task.cancel()
        if self.message_handler_task:
            self.message_handler_task.cancel()

    async def heartbeat(self):
        while self.ws and not self.ws.closed:
            try:
                logger.debug("Sending heartbeat")
                await self.ws.ping()

                # Log pending requests
                async with self._lock:
                    pending_requests = list(self._pending_requests.keys())
                logger.debug(f"Current pending requests: {pending_requests}")

                await asyncio.sleep(30)
            except Exception:
                logger.exception("Error in heartbeat")
                break

    async def message_handler(self):
        while True:
            try:
                message = await self.ws.recv()
                asyncio.create_task(self._process_message(message))
            except (ConnectionClosedError, ConnectionClosedOK):
                logger.warning("Connection to the gateway was closed.")
                await self.close_connection()
                break
            except Exception:
                logger.exception(f"Error receiving message")
                await self.close_connection()
                break

    async def _process_message(self, message: str):
        async with self._semaphore:
            try:
                data = json.loads(message)
                message_type = data.get("type")
                correlation_id = data.get("correlation_id")
                logger.debug(
                    textwrap.dedent(
                        f"""
                        Received message.
                        Type: {message_type}
                        Correlation ID: {correlation_id}
                    """
                    ).strip()
                )

                if correlation_id and self._pending_requests.get(correlation_id):
                    # Associate the response with the pending request
                    await self._handle_correlated_response(correlation_id, data)
                elif message_type in self.message_handlers:
                    await self.message_handlers[message_type](data)

                    if message_type in self.relevant_message_types():
                        # Trigger the timeout reset event for all pending requests
                        for timeout_event in self._timeout_reset_events.values():
                            timeout_event.set()
                else:
                    logger.warning(f"Unknown message type: {message_type}")

                # Optionally handle reset events or other logic here
                if message_type in self.relevant_message_types():
                    logger.debug(f"Ignoring reset_event for {message_type}")

            except Exception:
                logger.exception("Error processing message")

    def relevant_message_types(self) -> List[str]:
        """
        Define which message types should reset the timeout.
        Add all relevant message types that should reset the timeout here.
        """
        return [
            WsLocalTask.GENERATE.value,
            WsLocalTask.EVALUATE.value,
            # Add other message types as needed
        ]

    async def send_message(self, message: Dict[str, Any]):
        if self.ws and not self.ws.closed:
            logger.debug(
                textwrap.dedent(
                    f"""
                    Sending message.
                    Type: {message.get('type')}
                    Correlation ID: {message.get('correlation_id')}
                """
                ).strip()
            )
            await self.ws.send(json.dumps(message, cls=CustomJSONEncoder))
            logger.debug(
                textwrap.dedent(
                    f"""
                    Message sent.
                    Type: {message.get('type')}
                    Correlation ID: {message.get('correlation_id')}
                """
                ).strip()
            )
        else:
            raise ConnectionError("WebSocket connection is not established or closed")

    async def _handle_correlated_response(
        self, correlation_id: str, data: Dict[str, Any]
    ):
        if not correlation_id:
            logger.error("Received message without correlation_id. Ignoring.")
            return

        logger.debug(
            f"Handling correlated response for correlation_id: {correlation_id}"
        )
        async with self._lock:
            response_queue = self._responses.get(correlation_id)
            if response_queue:
                logger.debug(
                    f"Putting data into response queue for correlation_id: {correlation_id}"
                )
                await response_queue.put(data)
                event = self._pending_requests.get(correlation_id)
                if event:
                    logger.debug(f"Setting event for correlation_id: {correlation_id}")
                    event.set()
            else:
                logger.warning(
                    f"No response queue found for correlation_id: {correlation_id}"
                )
                logger.debug(
                    f"Current pending requests: {list(self._pending_requests.keys())}"
                )
                logger.debug(f"Current response queues: {list(self._responses.keys())}")

    async def _generate_correlation_id(self) -> str:
        return str(uuid4())

    async def _register_request(self, correlation_id: str):
        async with self._lock:
            event = asyncio.Event()
            self._pending_requests[correlation_id] = event
            self._timeout_reset_events[correlation_id] = asyncio.Event()
            self._responses[correlation_id] = (
                asyncio.Queue()
            )  # Ensure a queue is created
            logger.debug(f"Registered request with correlation_id: {correlation_id}")

    async def _unregister_request(self, correlation_id: str):
        async with self._lock:
            self._pending_requests.pop(correlation_id, None)
            self._responses.pop(correlation_id, None)
            self._timeout_reset_events.pop(correlation_id, None)
            logger.debug(f"Unregistered request with correlation_id: {correlation_id}")

    async def wait_for_response(
        self, correlation_id: str, timeout: float = 600.0
    ) -> Dict[str, Any]:
        async with self._lock:
            if correlation_id not in self._pending_requests:
                raise ValueError(
                    f"No pending request with correlation_id {correlation_id}."
                )
            event = self._pending_requests[correlation_id]
            timeout_event = self._timeout_reset_events[correlation_id]

        end_time = time.time() + timeout
        while True:
            remaining = end_time - time.time()
            if remaining <= 0:
                logger.error(
                    f"Timeout waiting for response. Correlation ID: {correlation_id}"
                )
                await self._unregister_request(correlation_id)
                raise asyncio.TimeoutError

            try:
                # Check if the WebSocket is closed
                if self.ws.closed:
                    logger.error("WebSocket connection is closed")
                    await self._unregister_request(correlation_id)
                    raise ConnectionError("WebSocket connection is closed")

                # Create tasks
                wait_tasks = [
                    asyncio.create_task(event.wait()),
                    asyncio.create_task(timeout_event.wait()),
                    asyncio.create_task(
                        self.ws.wait_closed()
                    ),  # New task to detect WebSocket closure
                ]

                done, pending = await asyncio.wait(
                    wait_tasks,
                    timeout=remaining,
                    return_when=asyncio.FIRST_COMPLETED,
                )

                # Cancel any pending tasks
                for task in pending:
                    task.cancel()

                if wait_tasks[0] in done:
                    response = await self._responses[correlation_id].get()
                    await self._unregister_request(correlation_id)
                    return response["data"]

                if wait_tasks[1] in done:
                    # Reset the timeout countdown
                    end_time = time.time() + timeout
                    timeout_event.clear()
                    logger.debug(f"Timeout reset for correlation_id: {correlation_id}")

                if wait_tasks[2] in done:
                    logger.error(
                        "WebSocket connection closed while waiting for response"
                    )
                    await self._unregister_request(correlation_id)
                    raise ConnectionError(
                        "WebSocket connection closed while waiting for response"
                    )

            except Exception:
                logger.exception("Error waiting for response")
                await self._unregister_request(correlation_id)
                raise

    async def request(self, type: WsServerTask, data: Dict[str, Any] = {}):
        if not self.ws or self.ws.closed:
            await self.connect_to_gateway()

        correlation_id = await self._generate_correlation_id()
        await self._register_request(correlation_id)

        message = {
            "correlation_id": correlation_id,
            "type": type.value,
            "data": data,
        }
        try:
            await self.send_message(message)
            logger.debug(f"Sent request to server. Message: {message}")

            response = await self.wait_for_response(correlation_id)
            logger.debug(f"Received response: {response}")
            return response
        except asyncio.TimeoutError:
            logger.error(f"Timeout waiting for response. Message: {message}")
            raise
        except ConnectionError as e:
            logger.error(f"WebSocket connection error: {str(e)}")
            await self.close_connection()
            raise
        except Exception as e:
            logger.exception(f"Error for request to server. Message: {message}")
            await self.close_connection()
            raise
        finally:
            await self._unregister_request(correlation_id)

    def register_handlers(self, handler_class):
        for attr_name in dir(handler_class):
            attr = getattr(handler_class, attr_name)
            if hasattr(attr, "message_type"):
                self.message_handlers[attr.message_type] = attr

        logger.info(f"Registered message handlers: {self.message_handlers}")
